"""streamparse.ext.invoke

invoke tasks that are imported into a project's task.py.

Should be used like this::

    from invoke import task, run
    from streamparse.ext.invoke import *

    # your other tasks
"""
from __future__ import absolute_import
from __future__ import print_function

from invoke import run, task

__all__ = ["stormlocal"]

def stormdeps(topology=None):
    print("Storm dependencies (via lein):")
    run("lein deps :tree")
    print("Python dependencies (via pip):")
    run("cat virtualenvs/{}".format(topology))

@task
def stormlocal(topology=None, time="5000", debug=False):
    if topology is None:
        print("Must specify topology for stormlocal")
        return
    run("mkdir -p _resources/resources")
    run("cp src/*.py _resources/resources/")
    cmd = "lein run -s {} -t {}".format(topology, time)
    if debug:
        cmd += " -d"
    run(cmd)



from __future__ import print_function

from .util import read_handshake, read_command, send_message, Component, json
from .util import _stdout


class Spout(Component):

    def initialize(self, storm_conf, context):
        pass

    def ack(self, tup_id):
        """Called when a bolt acknowledges a tuple in the topology.

        :param tup_id: a ``str`` representing the tuple that was acknowledged.
        """
        pass

    def fail(self, tup_id):
        """Called when a tuple fails in the topology, a Spout can choose to
        emit the tuple again or ignore the fail.  Default is to ignore.

        :param tup_id: a ``str`` representing the tuple that was failed.
        """
        pass

    def next_tuple(self):
        raise NotImplementedError()

    def emit(self, tup, tup_id=None, stream=None, direct_task=None):
        """Emit a spout tuple message.

        :param tup: a ``list`` representing the tuple to send to Storm, should
                    contain only JSON-serializable data.
        :param tup_id: ``str`` or ``int`` which is the id for the tuple. Leave
                    this blank for an unreliable emit.
        :param stream: ``str`` ID of the stream this tuple should be emitted
                    to.  Leave empty to emit to the default stream.
        :param direct_task: ``int`` indicating the task to send the tuple to if
                    performing a direct emit.
        """
        if not isinstance(tup, list):
            raise TypeError('All tuples must be lists, received {!r} instead'\
                .format(type(tup)))

        msg = {'command': 'emit', 'tuple': tup}
        if tup_id is not None:
            msg['id'] = tup_id
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        send_message(msg)

    def emit_many(self, tuples, tup_id=None, stream=None, direct_task=None):
        """A more efficient way to send many tuples, dumps out all tuples to
        STDOUT instead of writing one at a time.
        """
        msg = {
            'command': 'emit',
        }
        if tup_id is not None:
            msg['id'] = tup_id
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        lines = []
        for tup in tuples:
            msg['tuple'] = tup
            lines.append(json.dumps(msg))
            lines.append('end')
        print('\n'.join(lines), file=_stdout)

    def run(self):
        storm_conf, context = read_handshake()
        try:
            self.initialize(storm_conf, context)
            while True:
                cmd = read_command()
                if cmd['command'] == 'next':
                    self.next_tuple()
                if cmd['command'] == 'ack':
                    self.ack(cmd['id'])
                if cmd['command'] == 'fail':
                    self.fail(cmd['id'])
                send_message({'command': 'sync'})
        except Exception as e:
            self.raise_exception(e)



from __future__ import print_function

from .util import read_handshake, read_tuple, send_message, Component, json
from .util import _stdout


_ANCHOR_TUPLE = None

class Bolt(Component):

    def initialize(self, storm_conf, context):
        """Process initialization after Storm configuration and topology
        context are received from the worker process.
        """
        pass

    def process(self, tup):
        """Main method for tuple processing, must be overridden by subclasses.

        :param tup: a ``Tuple`` object.
        """
        raise NotImplementedError()

    def emit(self, tup, stream=None, anchors=[], direct_task=None):
        """Emit a bolt tuple message.

        :param tup: a ``list`` representing the tuple ot send to Storm, should
                    contain only JSON-serializable data.
        :param stream: a ``str`` indicating the ID of the stream to emit this
                       tuple to.  Leave empty to emit to default stream.
        :param anchors: a ``list`` which is the IDs of the tuples this output
                        tuple should be anchored to.
        :param direct_task: an ``int`` which indicates the task to send the
                            tuple to.
        """
        if not isinstance(tup, list):
            raise TypeError('All tuples must be lists, received {!r} instead'\
                            .format(type(tup)))
        if _ANCHOR_TUPLE is not None:
            anchors = [_ANCHOR_TUPLE]
        msg = {'command': 'emit', 'tuple': tup}
        if stream is not None:
            msg['stream'] = stream
        msg['anchors'] = [x.id for x in anchors]
        if direct_task is not None:
            msg['task'] = direct_task

        send_message(msg)

    def emit_many(self, tuples, stream=None, anchors=[], direct_task=None):
        """A more efficient way to send many tuples, dumps out all tuples to
        STDOUT instead of writing one at a time.
        """
        if _ANCHOR_TUPLE is not None:
            anchors = [_ANCHOR_TUPLE]
        msg = {
            'command': 'emit',
            'anchors': [a.id for a in anchors],
        }
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        lines = []
        for tup in tuples:
            msg['tuple'] = tup
            lines.append(json.dumps(msg))
            lines.append('end')
        print('\n'.join(lines), file=_stdout)

    def ack(self, tup):
        """Acknowledge a tuple.

        :param tup: a ``Tuple`` object.
        """
        send_message({'command': 'ack', 'id': tup.id})

    def fail(self, tup):
        """Fail a tuple.

        :param tup: a ``Tuple`` object.
        """
        send_message({'command': 'fail', 'id': tup.id})

    def run(self):
        """Main run loop for all bolts, performs initial handshake and
        reads tuples handing them off to the subclass.  Any exceptions are
        caught and sent back to Storm as a proper exception.
        """
        storm_conf, context = read_handshake()
        try:
            self.initialize(storm_conf, context)
            while True:
                tup = read_tuple()
                self.process(tup)
        except Exception as e:
            self.raise_exception(e)


class BasicBolt(Bolt):

    def run(self):
        global _ANCHOR_TUPLE
        storm_conf, context = read_handshake()
        try:
            self.initialize(storm_conf, context)
            while True:
                tup = read_tuple()
                _ANCHOR_TUPLE = tup
                self.process(tup)
                self.ack(tup)
        except Exception as e:
            self.raise_exception(e)




from docopt import docopt
from invoke import run

import json

from os import path

from bootstrap import quickstart


# XXX: these are commands we're working on still
TODO_CMDS = """
        sparse setup [-e <env>]
        sparse debug [-e <env>]
        sparse kill [-e <env>]
        sparse restart [-e <env>]
        sparse attach [-e <env>]
        sparse logs [-e <env>]

"""


def main():
    """sparse: manage streamparse clusters.

    sparse provides a front-end to streamparse, a framework for creating Python
    projects for running, debugging, and submitting computation topologies against
    real-time streams, using Apache Storm.

    It requires the lein (Clojure build tool) to be on your $PATH, and uses
    lein and Clojure under the hood for JVM interop.

    Usage:
        sparse quickstart <project_dir>
        sparse run [-e <env>] [-t <time>] [--debug]
        sparse deploy [-e <env>]
        sparse list
        sparse (-h | --help)
        sparse --version

    Options:
        -h --help         Show this screen.
        --version         Show version.
        -e <env>          Set environment; as described in config.json [default: local].
        -t <time>         Time (in milliseconds) to keep cluster running [default: 10000].
        --verbose         Verbose output.
        --debug           Debug output.
    """
    args = docopt(main.__doc__, version="sparse 0.1")
    if args["run"]:
        cfg = json.load(open("config.json"))
        topo_dir = cfg["topology_specs"]
        topo_first = cfg["topologies"][0]
        print "Running {topo_name} topology...".format(topo_name=topo_first)
        time = args["-t"]
        debug = args["--debug"]
        topology = path.join(topo_dir, topo_first) + ".clj"
        cmd = "invoke stormlocal --topology={topology} --time={time}".format(
            topology=topology, time=time, debug=debug)
        if debug:
            cmd += " --debug"
        run(cmd)
    elif args["list"]:
        print "invoke (local) tasks:"
        run("invoke -l")
        print
        print "fabric (remote) tasks:"
        run("fab -l")
    elif args["quickstart"]:
        quickstart(args['<project_dir>'])


if __name__ == "__main__":
    main()



"""streamparse.ext.fabric

fabric tasks that are imported into a project's fabfile.py.

Should be used like this::

    from fabric.api import *
    from streamparse.ext.fabric import *

    # your other tasks
"""
from __future__ import absolute_import
import json
import os

from fabric.api import *
from .util import get_config

# __all__ = ["setup_virtualenv", "workers"]
__all__ = ['env', 'deploy_topology', 'uberjar', 'deploy_virtualenv',
           'submit_topology']


@task
def env(env_=None):
    """Activate a particular environment from the config.json file."""
    with open('config.json', 'r') as fp:
        config = json.load(fp)
    _env.nimbus = config['envs'][env_]['hosts']['nimbus']
    _env.workers = config['envs'][env_]['hosts']['workers']


@task
def deploy_topology(topology=None):
    """Deploy a topology to a remote host.  Deploying a streamparse topology
    accomplishes two things:
    1. Create an uberjar which contains all code.
    2. Push the topology virtualenv requirements to remote.
    3. Update virtualenv on host server.
    4. Submit topology (in uberjar) to remote Storm cluster."""
    pass


@task
def uberjar():
    """Create a JAR which contains all required dependencies for execution on a
    Storm cluster."""
    pass


@task
def deploy_virtualenv():
    """Deploy and update the virtualenv on all Storm worker nodes."""
    pass

@task
def submit_topology(topology_name, timeout=None):
    """Submit a Storm topology to the Nimbus machine in a Storm cluster."""
    pass



"""
Both ends of this protocol use a line-reading mechanism, so be sure to trim off newlines from the input and to append them to your output.
All JSON inputs and outputs are terminated by a single line containing "end". Note that this delimiter is not itself JSON encoded.
The bullet points below are written from the perspective of the script writer's STDIN and STDOUT.
"""
from __future__ import print_function
try:
    import simplejson as json
except ImportError:
    import json
import logging
import os
import sys
from collections import deque
import traceback


config = context = None
storm_log = logging.getLogger('pystorm')

_MAX_MESSAGE_SIZE = 16777216
_MAX_BLANK_MSGS = 500
_MAX_LINES = 100
# queue up commands we read while trying to read task IDs
_pending_commands = deque()
# queue up task IDs we read while trying to read commands/tuples
_pending_task_ids = deque()
# we'll redirect stdout, but we'll save original for communication to Storm
_stdout = sys.stdout


class StormIPCException(Exception): pass


class LogStream(object):
    """Object that implements enough of the Python stream API to be used as
    sys.stdout and sys.stderr. Messages are written to the Python logger.
    """
    def __init__(self, logger):
        self.logger = logger

    def write(self, message):
        for line in message.split('\n'):
            self.logger.error(line)


## Message recieving

def read_message_lines():
    lines = blank_lines = 0

    while True:
        line = sys.stdin.readline()[0:-1]  # ignore new line
        if line == 'end':
            break
        lines += 1
        message_size = len(line)

        # If message size exceeds MAX_MESSAGE_SIZE, we assume that the
        # Storm worker has died, and we would be reading an infinite
        # series of blank lines. Throw an error to halt processing,
        # otherwise the task will use 100% CPU and will quickly consume
        # a huge amount of RAM.
        if message_size >= _MAX_MESSAGE_SIZE:
            raise StormIPCException(('Message {} exceeds '
                                     '{:,}'.format(line, _MAX_MESSAGE_SIZE)))

        # If Storm starts to send us a series of blank lines, after awhile we
        # have to assume that the pipe to the Storm supervisor is broken, this
        # is one of the more annoying parts of Petrel and we should probably
        # tune _MAX_BLANK_MSGS
        if line == '':
            blank_lines += 1
            if blank_lines >= _MAX_BLANK_MSGS:
                raise StormIPCException(('{:,} blank lines received, assuming '
                                         'pipe to Storm supervisor is '
                                         'broken'.format(blank_lines)))

        if lines >= _MAX_LINES:
            raise StormIPCException(('Message exceeds {:,} lines, assuming this'
                                     'is an error.'.format(lines)))

        yield line


def read_message():
    """Read a message from Storm, reconstruct newlines appropriately."""
    message = ''.join('{}\n'.format(line) for line in read_message_lines())
    return json.loads(message)


def read_task_ids():
    if _pending_task_ids:
        return _pending_task_ids.popleft()
    else:
        msg = read_message()
        while not isinstance(msg, list):
            _pending_commands.append(msg)
            msg = read_message()
        return msg


def read_command():
    if _pending_commands:
        return _pending_commands.popleft()
    else:
        msg = read_message()
        while isinstance(msg, list):
            _pending_task_ids.append(msg)
            msg = read_message()
        return msg


def read_tuple():
    cmd = read_command()
    return Tuple(cmd['id'], cmd['comp'], cmd['stream'], cmd['task'], cmd['tuple'])


def read_handshake():
    """Read and process an initial handshake message from Storm."""
    # Redirect stdout and stderr to ensure that print statements/functions
    # won't crash the Storm Java worker
    sys.stdout = LogStream(logging.getLogger('pystorm.stdout'))
    sys.stderr = LogStream(logging.getLogger('pystorm.stderr'))

    msg = read_message()
    storm_log.info('Received initial handshake from Storm: %r', msg)
    # Write a blank PID file out to the pidDir
    pid_dir = msg['pidDir']
    pid = os.getpid()
    open('{}/{}'.format(pid_dir, str(pid)), 'w').close()
    send_message({'pid': pid})
    storm_log.info('Process ID sent to Storm')

    return msg['conf'], msg['context']


## Message sending

def send_message(message):
    """Send a message to Storm via stdout"""
    print(json.dumps(message), file=_stdout)
    print('end', file=_stdout)
    _stdout.flush()


class Component(object):
    """Base class for Spouts and Bolts which contains class methods for
    logging messages back to the Storm worker process.
    """

    def raise_exception(self, exception):
        """Report an exception back to Storm.

        :param exception: a Python exception.
        """
        self.log('Python exception raised: {}\n{}'\
                 .format(exception.__class__.__name__,
                         traceback.format_exc(exception)))
        send_message({'command': 'sync'})  # sync up right away

    def log(self, message, level='info'):
        """Log a message to Storm optionally providing a logging level.

        :param message: any object that supports ``__str__()``.
        :param level: a ``str`` representing the log level.
        """
        send_message({'command': 'log', 'msg': str(message), 'level': level})


class Tuple(object):

    __slots__ = ['id', 'component', 'stream', 'task', 'values']

    def __init__(self, id, component, stream, task, values):
        self.id = id
        self.component = component
        self.stream = stream
        self.task = task
        self.values = values

    def __repr__(self):
        return '<{}: {}> {}'.format(self.__class__.__name__, self.id,
                                    self.values)

